package org.jboss.qa.brms;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response.Status;
import org.apache.commons.net.util.Base64;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.CredentialsProvider;
import org.apache.http.client.HttpClient;
import org.apache.http.impl.client.BasicCredentialsProvider;
import org.apache.http.impl.client.HttpClientBuilder;
import org.codehaus.jackson.map.ObjectMapper;
import org.guvnor.rest.client.AddRepositoryToOrganizationalUnitRequest;
import org.guvnor.rest.client.CompileProjectRequest;
import org.guvnor.rest.client.CreateOrCloneRepositoryRequest;
import org.guvnor.rest.client.CreateOrganizationalUnitRequest;
import org.guvnor.rest.client.CreateProjectRequest;
import org.guvnor.rest.client.DeleteProjectRequest;
import org.guvnor.rest.client.DeployProjectRequest;
import org.guvnor.rest.client.InstallProjectRequest;
import org.guvnor.rest.client.JobRequest;
import org.guvnor.rest.client.JobResult;
import org.guvnor.rest.client.JobStatus;
import org.guvnor.rest.client.OrganizationalUnit;
import org.guvnor.rest.client.ProjectRequest;
import org.guvnor.rest.client.ProjectResponse;
import org.guvnor.rest.client.RemoveOrganizationalUnitRequest;
import org.guvnor.rest.client.RemoveRepositoryFromOrganizationalUnitRequest;
import org.guvnor.rest.client.RemoveRepositoryRequest;
import org.guvnor.rest.client.RepositoryRequest;
import org.guvnor.rest.client.RepositoryResponse;
import org.jboss.resteasy.client.ClientExecutor;
import org.jboss.resteasy.client.ClientRequest;
import org.jboss.resteasy.client.ClientRequestFactory;
import org.jboss.resteasy.client.ClientResponse;
import org.jboss.resteasy.client.core.BaseClientResponse;
import org.jboss.resteasy.client.core.executors.ApacheHttpClient4Executor;
import org.jboss.resteasy.spi.BadRequestException;
import org.jboss.resteasy.spi.ReaderException;
import org.jboss.resteasy.spi.ResteasyProviderFactory;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentDescriptor;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentJobResult;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class RestWorkbenchClient implements WorkbenchClient {

    private static final Logger LOG = LoggerFactory.getLogger(RestWorkbenchClient.class);
    private static final int GET = 0;
    private static final int POST = 1;
    private static final int DELETE = 2;
    private final String AUTH_HEADER;
    private final String appUrl;
    private final String userId;
    private final String password;

    private <T extends Object> T get(ClientRequest request, Class<T> returnType) {
        return process(request, GET, returnType);
    }

    private <T extends Object> T post(ClientRequest request, Class<T> returnType) {
        return process(request, POST, returnType);
    }

    private <T extends Object> T delete(ClientRequest request, Class<T> returnType) {
        return process(request, DELETE, returnType);
    }

    private <T extends Object> T process(ClientRequest request, int method, Class<T> returnType) {
        ClientResponse<T> response = null;
        try {
            switch (method) {
            case GET:
                response = request.get(returnType);
                break;
            case POST:
                response = request.post(returnType);
                break;
            case DELETE:
                response = request.delete(returnType);
                break;
            default:
                throw new AssertionError();
            }
        } catch (Exception ex) {
            throw new RequestFailureException(ex);
        }
        try {
            int statusCode = response.getStatus();
            if (statusCode == 404) {
                throw new ResponseFailureException("Requested resource does not exist.");
            }
            String contentType = response.getHeaders().getFirst("Content-Type");
            if (MediaType.APPLICATION_JSON.equals(contentType) || MediaType.APPLICATION_XML.equals(contentType)) {
                try {
                    checkResponse(response);
                } catch (Exception ex) { // TODO add throws to all the
                                         // operations
                    ex.printStackTrace();
                    return null;
                }
                T res = response.getEntity();
                return res;
            }
            // now that we know that the result is wrong, try to identify the
            // reason
            final String responseBody = response.getEntity(String.class).replaceAll("><", ">\n<");
            final String headerTag = "<h1>";
            final int h1StartPosition = responseBody.indexOf(headerTag);
            if (h1StartPosition < 0) {
                LOG.error("Failed processing REST request. Full response body on DEBUG.");
                LOG.debug("Response body: {}", responseBody);
                throw new ResponseFailureException("Unexpected content-type: " + contentType);
            } else {
                final int h1EndPosition = responseBody.indexOf("</h1>");
                final String reason = responseBody.substring(h1StartPosition + headerTag.length(), h1EndPosition);
                LOG.error("Failed processing REST request, reason given: '{}'. Full response body on DEBUG.", reason);
                LOG.debug("Response body: {}", responseBody);
                throw new ResponseFailureException("Failed processing REST request, reson given: " + reason);
            }
        } catch (ReaderException ex) {
            response.resetStream();
            LOG.error("Bad entity: {}", response.getEntity(String.class));
            throw new ResponseFailureException(ex);
        } finally {
            response.releaseConnection();
        }
    }

    private void checkResponse(ClientResponse<?> responseObj) throws Exception {
        ClientResponse<?> test = BaseClientResponse.copyFromError(responseObj);
        responseObj.resetStream();
        if (test.getResponseStatus() == Status.BAD_REQUEST) {
            throw new BadRequestException(test.getEntity(String.class));
        } else if (test.getResponseStatus() != Status.OK && test.getResponseStatus() != Status.ACCEPTED && test.getResponseStatus() != Status.NOT_FOUND) {
            throw new Exception("Request operation failed. Response status = " + test.getResponseStatus() + "\n\n"
                    + test.getEntity(String.class));
        } else {
            LOG.info("Response entity: [{}]", test.getEntity(String.class));
        }
    }

    public RestWorkbenchClient(String appUrl, String userId, String password) {
        this.appUrl = appUrl;
        this.userId = userId;
        this.password = password;
        this.AUTH_HEADER = "Basic "
                + Base64.encodeBase64String(String.format("%s:%s", userId, password).getBytes()).trim();
    }

    public static RestWorkbenchClient createInstance(String appUrl, String userId, String password) {
        return new RestWorkbenchClient(appUrl, userId, password);
    }

    @Override
    public JobResult checkJobStatus(String jobId) {
        return get(createRequest(appUrl + "/rest/jobs/" + jobId), JobResult.class);
    }

    @Override
    public JobRequest deleteRepository(String repo) {
        LOG.info("Deleting repository '{}'", repo);
        RemoveRepositoryRequest entity = delete(createRequest(appUrl + "/rest/repositories/" + repo),
                RemoveRepositoryRequest.class);
        if (entity.getStatus() == JobStatus.ACCEPTED || entity.getStatus() == JobStatus.SUCCESS || entity.getStatus() == JobStatus.APPROVED) {
            return entity;
        } else {
            throw new BadResponseStatusException(entity);
        }
    }

    @Override
    public JobRequest createRepository(String repo, String cloneRepoUrl, String orgUnit) {
        LOG.info("Cloning repo '{}' from URL '{}'", repo, cloneRepoUrl);
        RepositoryRequest repoRequest = new RepositoryRequest();
        repoRequest.setName(repo);
        repoRequest.setRequestType("clone");
        repoRequest.setGitURL(cloneRepoUrl);
        repoRequest.setOrganizationalUnitName(orgUnit);
        String input = null;
        try {
            input = new ObjectMapper().writeValueAsString(repoRequest);
        } catch (Exception e) {
            throw new RuntimeException("Unable to serialize RepositoryRequest", e);
        }
        return post(createRequest(appUrl + "/rest/repositories/", input), CreateOrCloneRepositoryRequest.class);
    }

    @Override
    public RepositoryResponse getRepository(String repo) throws NotSuccessException {
        return get(createRequest(appUrl + "/rest/repositories/" + repo), RepositoryResponse.class);
    }

    @Override
    public JobRequest newRepository(String repo, String orgUnit) {
        LOG.info("Creating new repo '{}'", repo);
        RepositoryRequest repoRequest = new RepositoryRequest();
        repoRequest.setName(repo);
        repoRequest.setRequestType("new");
        repoRequest.setOrganizationalUnitName(orgUnit);
        String input = null;
        try {
            input = new ObjectMapper().writeValueAsString(repoRequest);
        } catch (Exception e) {
            throw new RuntimeException("Unable to serialize RepositoryRequest", e);
        }
        return post(createRequest(appUrl + "/rest/repositories/", input), CreateOrCloneRepositoryRequest.class);
    }
    
    

    @Override
    public Collection<RepositoryResponse> getRepositories() {
        // JAXB returns Collection<LinkedHashMap>, probably wrong
        // deserialization
        Collection<Map<String, String>> result = get(createRequest(appUrl + "/rest/repositories"), Collection.class);
        Collection<RepositoryResponse> repositoryResponse = new ArrayList<RepositoryResponse>();
        for (Map<String, String> map : result) {
            RepositoryResponse resp = new RepositoryResponse();
            resp.setName(map.get("name"));
            resp.setDescription(map.get("description"));
            resp.setGitURL(map.get("gitURL"));
            repositoryResponse.add(resp);
        }
        return repositoryResponse;
    }

    @Override
    public JobRequest createOrganizationalUnit(String name, String owner, String... repositories) {
        LOG.info("Creating organizational unit '{}' owned by '{}' containing [{}]", name, owner, repositories);
        OrganizationalUnit ou = new OrganizationalUnit();
        ou.setRepositories(new ArrayList<String>());
        for (int i = 0; repositories != null && i < repositories.length; ++i) {
            ou.getRepositories().add(repositories[i]);
        }
        ou.setName(name);
        ou.setOwner(owner);
        String input = null;
        try {
            input = new ObjectMapper().writeValueAsString(ou);
        } catch (Exception e) {
            throw new RuntimeException("Unable to serialize " + ou.getClass().getSimpleName(), e);
        }
        return post(createRequest(appUrl + "/rest/organizationalunits/", input), CreateOrganizationalUnitRequest.class);
    }

    public JobRequest deleteOrganizationalUnit(String organizationalUnit) {
        LOG.info("Removing organizational unit '{}'", organizationalUnit);
        RemoveOrganizationalUnitRequest entity = delete(createRequest(appUrl + "/rest/organizationalunits/" + organizationalUnit),
                RemoveOrganizationalUnitRequest.class);
        if (entity.getStatus() == JobStatus.ACCEPTED || entity.getStatus() == JobStatus.SUCCESS || entity.getStatus() == JobStatus.APPROVED) {
            return entity;
        } else {
            throw new BadResponseStatusException(entity);
        }
    }

    @Override
    public Collection<OrganizationalUnit> getOrganizationalUnits() throws NotSuccessException {
        Collection<Map<String, String>> result = get(createRequest(appUrl + "/rest/organizationalunits"), Collection.class);
        Collection<OrganizationalUnit> list = new ArrayList<OrganizationalUnit>();
        for (Map<String, String> map : result) {
            OrganizationalUnit ou = new OrganizationalUnit();
            ou.setName(map.get("name"));
            ou.setDescription(map.get("description"));
            ou.setOwner(map.get("owner"));
            ou.setRepositories(null); // TODO!
            list.add(ou);
        }
        return list;
    }

    @Override
    public OrganizationalUnit getOrganizationalUnit(String organizationalUnit) throws NotSuccessException {
        return get(createRequest(appUrl + "/rest/organizationalunits/" + organizationalUnit), OrganizationalUnit.class);
    }

    @Override
    public JobRequest addRepositoryToOrganizationalUnit(String organizationalUnit, String repo) {
        LOG.info("Adding repo '{}' to organizational unit '{}'", repo, organizationalUnit);
        return post(
                createRequest(appUrl + "/rest/organizationalunits/" + organizationalUnit + "/repositories/" + repo),
                AddRepositoryToOrganizationalUnitRequest.class);
    }

    @Override
    public JobRequest removeRepositoryFromOrganizationalUnit(String organizationalUnit, String repo) {
        LOG.info("Removing repo '{}' from organizational unit '{}'", repo, organizationalUnit);
        return delete(createRequest(appUrl + "/rest/organizationalunits/" + organizationalUnit + "/repositories/"
                + repo), RemoveRepositoryFromOrganizationalUnitRequest.class);
    }

    @Override
    public JobRequest removeOrganizationalUnit(String organizationalUnit) {
        LOG.info("Removing group '{}'", organizationalUnit);
        return delete(createRequest(appUrl + "/rest/organizationalunits/" + organizationalUnit), RemoveOrganizationalUnitRequest.class);
    }

    @Override
    public JobRequest createProject(String repo, String project, String groupId, String version, String description) throws NotSuccessException {
        ProjectRequest req = new ProjectRequest();
        req.setName(project);
        req.setGroupId(groupId);
        req.setVersion(version);
        req.setDescription(description);
        
        String input = null;
        try {
            input = new ObjectMapper().writeValueAsString(req);
        } catch (Exception e) {
            throw new RuntimeException("Unable to serialize " + req.getClass().getSimpleName(), e);
        }
        System.out.println(input);
        return post(createRequest(appUrl + "/rest/repositories/" + repo + "/projects", input), CreateProjectRequest.class);
    }
    
    @Override
    public Collection<ProjectResponse> getAllProjects(String repo) {
        Collection<Map<String, String>> result = get(createRequest(appUrl + "/rest/repositories/" + repo + "/projects"), Collection.class);
        Collection<ProjectResponse> list = new ArrayList<ProjectResponse>();
        for (Map<String, String> map : result) {
            ProjectResponse p = new ProjectResponse();
            p.setName(map.get("name"));
            p.setGroupId(map.get("groupId"));
            p.setVersion(map.get("version"));
            p.setDescription(map.get("description"));
            list.add(p);
        }
        return list;
    }

    @Override
    public JobRequest deleteProject(String repo, String project) throws NotSuccessException {
        LOG.info("Removing project '{}' from repo '{}'", project, repo);
        return delete(createRequest(appUrl + "/rest/repositories/" + repo + "/projects/" + project), DeleteProjectRequest.class);
    }

    @Override
    public JobRequest testProject(String repo, String project) throws NotSuccessException {
        LOG.info("Testing project '{}' from repo '{}'", project, repo);
        return post(maven(repo, project, "test"), CompileProjectRequest.class);
    }

    @Override
    public JobRequest compileProject(String repo, String project) {
        LOG.info("Compiling project '{}' from repo '{}'", project, repo);
        return post(maven(repo, project, "compile"), CompileProjectRequest.class);
    }

    @Override
    public JobRequest installProject(String repo, String project) {
        LOG.info("Installing project '{}' from repo '{}'", project, repo);
        return post(maven(repo, project, "install"), InstallProjectRequest.class);
    }

    @Override
    public JobRequest deployProject(String repo, String project) {
        LOG.info("Deploying project '{}' from repo '{}'", project, repo);
        return post(maven(repo, project, "deploy"), DeployProjectRequest.class);
    }

    private ClientRequest maven(String repo, String project, String operation) {
        return createRequest(appUrl + "/rest/repositories/" + repo + "/projects/" + project + "/maven/" + operation);
    }

    @Override
    public JaxbDeploymentJobResult createDeploymentUnit(String deploymentId, String strategy)
            throws NotSuccessException {
        return createDeploymentUnit(deploymentId, strategy, null, null);
    }

    @Override
    public JaxbDeploymentJobResult createDeploymentUnit(String deploymentId, String strategy, String kBase,
            String kSession) throws NotSuccessException {
        StringBuilder builder = new StringBuilder(appUrl);
        builder.append("/rest/deployment/").append(deploymentId);
        if (kBase != null && !kBase.isEmpty()) {
            builder.append(":").append(kBase);
        }
        if (kSession != null && !kSession.isEmpty()) {
            builder.append(":").append(kSession);
        }
        builder.append("/deploy");
        if (strategy != null && !strategy.isEmpty()) {
            builder.append("?strategy=").append(strategy);
        }

        // Has to send empty descriptor because Weblogic Jersey 1.x marshalls injected parameter and if we send empty body then it crashes
        JaxbDeploymentJobResult jr = post(createRequest(builder.toString(), MediaType.APPLICATION_XML_TYPE, new JaxbDeploymentDescriptor()),
                JaxbDeploymentJobResult.class); // JSON does not work due to BZ 1050894
        if (!jr.isSuccess()) {
            throw new NotSuccessException(null, "The deployment unit was not created successfully.");
        }

        return jr;
    }

    @Override
    public JaxbDeploymentUnit getDeploymentUnit(String deploymentId) throws NotSuccessException {
        JaxbDeploymentUnit ret = null;
        ClientRequest req = createRequest(appUrl + "/rest/deployment/" + deploymentId, MediaType.APPLICATION_XML_TYPE);
        ret = get(req, JaxbDeploymentUnit.class);
        return ret;
    }

    @Override
    public JaxbDeploymentJobResult removeDeploymentUnit(String deploymentId) throws NotSuccessException {
        return post(
                createRequest(appUrl + "/rest/deployment/" + deploymentId + "/undeploy",
                        MediaType.APPLICATION_XML_TYPE), JaxbDeploymentJobResult.class);
    }

    private ClientRequest createRequest(String resourcePath, String body) {
        return createRequest(resourcePath).body(MediaType.APPLICATION_JSON_TYPE, body);
    }

    private ClientRequest createRequest(String resourcePath) {
        return getClientRequestFactory().createRequest(resourcePath).header("Authorization", AUTH_HEADER)
                .accept(MediaType.APPLICATION_JSON_TYPE).followRedirects(true);
    }

    private ClientRequest createRequest(String resourcePath, MediaType mt) {
        return getClientRequestFactory().createRequest(resourcePath).header("Authorization", AUTH_HEADER).accept(mt)
                .followRedirects(true);
    }

    private ClientRequest createRequest(String resourcePath, MediaType mt, Object body) {
        return createRequest(resourcePath, mt).body(mt, body);
    }

    private ClientRequestFactory getClientRequestFactory() {
        HttpClientBuilder builder = HttpClientBuilder.create();
        CredentialsProvider cp = new BasicCredentialsProvider();
        cp.setCredentials(new AuthScope(AuthScope.ANY_HOST, AuthScope.ANY_PORT, AuthScope.ANY_REALM),
                new UsernamePasswordCredentials(userId, password));
        HttpClient httpClient = builder.setDefaultCredentialsProvider(cp).build();
        ClientExecutor clientExecutor = new ApacheHttpClient4Executor(httpClient);
        return new ClientRequestFactory(clientExecutor, ResteasyProviderFactory.getInstance());
    }

}
