package org.jboss.qa.brms;

import java.util.Collection;
import org.guvnor.rest.client.JobRequest;
import org.guvnor.rest.client.JobResult;
import org.guvnor.rest.client.JobStatus;
import org.guvnor.rest.client.OrganizationalUnit;
import org.guvnor.rest.client.ProjectResponse;
import org.guvnor.rest.client.RepositoryResponse;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentJobResult;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentUnit;
import org.kie.services.client.serialization.jaxb.impl.deploy.JaxbDeploymentUnit.JaxbDeploymentStatus;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class WorkbenchClientWithAcknowledgement implements WorkbenchClient {

    private static final Logger LOG = LoggerFactory.getLogger(WorkbenchClientWithAcknowledgement.class);
    private static final int JOB_TIMEOUT = 300;
    private static final int MAX_ATTEMPTS_DEPLOYMENT_UNIT = 30;
    private final WorkbenchClient wc;

    public WorkbenchClientWithAcknowledgement(WorkbenchClient wc) {
        this.wc = wc;
    }

    /**
     * Waiting for a job to be successfully finished.
     *
     * @param jobId job identifier
     * @return job result
     * @throws NotSuccessException in case final job status was different from SUCCESS
     */
    @Override
    public JobResult checkJobStatus(String jobId) throws NotSuccessException {

        // Wait for job to leave status ACCEPTED/APPROVED
        long start = System.currentTimeMillis();
        JobResult res = wc.checkJobStatus(jobId);
        while (!isFinished(res) && System.currentTimeMillis() - start < JOB_TIMEOUT * 1000) {
            sleep(1);
            res = wc.checkJobStatus(jobId);
        }

        if (!isFinished(res)) {
            throw new NotSuccessException(null, String.format(
                    "Job %s hasn't finished within %ds. Last known status is %s.",
                    jobId, JOB_TIMEOUT, res.getStatus()));
        }

        if ((res.getStatus() != JobStatus.SUCCESS)) {
            throw new NotSuccessException(res);
        }
        return res;
    }

    private boolean isFinished(JobResult jr) {
        return jr.getStatus() != JobStatus.ACCEPTED && jr.getStatus() != JobStatus.APPROVED;
    }

    /**
     * Waiting for a deployment to be successfully deployed or undeployed.
     *
     * @param deploymentId
     * @param status
     * @return
     * @throws NotSuccessException
     */
    public JaxbDeploymentUnit checkDeploymentUnit(String deploymentId, JaxbDeploymentStatus status)
            throws NotSuccessException {
        JaxbDeploymentUnit du = null;
        for (int i = 0; i < MAX_ATTEMPTS_DEPLOYMENT_UNIT; ++i) {
            try {
                du = wc.getDeploymentUnit(deploymentId);
            } catch (Exception ex) {
                if (ex.getMessage().contains("does not exist") || ex.getMessage().contains("Response status = Not Found")) {
                    throw new NotSuccessException(null, "Deployment does not exist.", ex);
                }
                // it can fail because the kjar is deployed and the log cannot
                // be found
                sleep(3);
                continue;
            }
            if (du == null) {
                continue;
            }
            if (du.getStatus() == status || du.getStatus() == JaxbDeploymentStatus.ACCEPTED) {
                sleep(3);
            } else {
                break;
            }
        }

        if (du.getStatus() == status) {
            throw new NotSuccessException(null, "Timeout after " + MAX_ATTEMPTS_DEPLOYMENT_UNIT + " attempts.");
        }

        if (du.getStatus() == JaxbDeploymentStatus.DEPLOY_FAILED) {
            throw new NotSuccessException(null, "Deploy failed.");
        }

        if (du.getStatus() == JaxbDeploymentStatus.UNDEPLOY_FAILED) {
            throw new NotSuccessException(null, "Undeploy failed.");
        }
        return du;
    }

    @Override
    public JobRequest deleteRepository(String repo) throws NotSuccessException {
        JobRequest jr = wc.deleteRepository(repo);
        if (jr == null) {
            throw new NotSuccessException(null, "The repository '" + repo
                    + "' was not deleted. Probably it didn't exist before.");
        }
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest createRepository(String repo, String cloneRepoUrl, String orgUnit) throws NotSuccessException {
        JobRequest jr = wc.createRepository(repo, cloneRepoUrl, orgUnit);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public RepositoryResponse getRepository(String repo) throws NotSuccessException {
        return wc.getRepository(repo);
    }

    @Override
    public JobRequest createOrganizationalUnit(String name, String owner, String... repositories)
            throws NotSuccessException {
        JobRequest jr = wc.createOrganizationalUnit(name, owner, repositories);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest deleteOrganizationalUnit(String organizationalUnit)
            throws NotSuccessException {
        JobRequest jr = wc.deleteOrganizationalUnit(organizationalUnit);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public Collection<OrganizationalUnit> getOrganizationalUnits() throws NotSuccessException {
        return wc.getOrganizationalUnits();
    }

    @Override
    public OrganizationalUnit getOrganizationalUnit(String organizationalUnit) throws NotSuccessException {
        return wc.getOrganizationalUnit(organizationalUnit);
    }

    @Override
    public JobRequest addRepositoryToOrganizationalUnit(String organizationalUnit, String repo)
            throws NotSuccessException {
        JobRequest jr = wc.addRepositoryToOrganizationalUnit(organizationalUnit, repo);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest removeRepositoryFromOrganizationalUnit(String organizationalUnit, String repo)
            throws NotSuccessException {
        JobRequest jr = wc.removeRepositoryFromOrganizationalUnit(organizationalUnit, repo);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest removeOrganizationalUnit(String organizationalUnit) throws NotSuccessException {
        JobRequest jr = wc.removeOrganizationalUnit(organizationalUnit);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest createProject(String repo, String project, String groupId, String version, String description) throws NotSuccessException {
        JobRequest jr = wc.createProject(repo, project, groupId, version, description);
        checkJobStatus(jr.getJobId());
        return jr;
    }
    
    @Override
    public Collection<ProjectResponse> getAllProjects(String repo) {
        return wc.getAllProjects(repo);
    }

    @Override
    public JobRequest deleteProject(String repo, String project) throws NotSuccessException {
        JobRequest jr = wc.deleteProject(repo, project);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest testProject(String repo, String project) throws NotSuccessException {
        JobRequest jr = wc.testProject(repo, project);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest compileProject(String repo, String project) throws NotSuccessException {
        JobRequest jr = wc.compileProject(repo, project);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest installProject(String repo, String project) throws NotSuccessException {
        JobRequest jr = wc.installProject(repo, project);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JobRequest deployProject(String repo, String project) throws NotSuccessException {
        JobRequest jr = wc.deployProject(repo, project);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public JaxbDeploymentJobResult createDeploymentUnit(String deploymentId, String strategy)
            throws NotSuccessException {
        return createDeploymentUnit(deploymentId, strategy, null, null);
    }

    @Override
    public JaxbDeploymentJobResult createDeploymentUnit(String deploymentId, String strategy, String kBase,
            String kSession) throws NotSuccessException {
        JaxbDeploymentJobResult jr = wc.createDeploymentUnit(deploymentId, strategy, kBase, kSession);
        checkDeploymentUnit(deploymentId, JaxbDeploymentStatus.DEPLOYING);
        return jr;
    }

    @Override
    public JaxbDeploymentUnit getDeploymentUnit(String deploymentId) throws NotSuccessException {
        return wc.getDeploymentUnit(deploymentId);
    }

    @Override
    public JaxbDeploymentJobResult removeDeploymentUnit(String deploymentId) throws NotSuccessException {
        JaxbDeploymentJobResult jr = null;
        try {
            jr = wc.removeDeploymentUnit(deploymentId);
            Thread.sleep(1000);
            checkDeploymentUnit(deploymentId, JaxbDeploymentStatus.UNDEPLOYING);
            Thread.sleep(2000);
        } catch (NotSuccessException ex) {
            if (!ex.getMessage().contains("Deployment does not exist")) {
                throw ex;
            }
        } catch (InterruptedException e) {
            
        }
        return jr;
    }

    @Override
    public JobRequest newRepository(String repo, String orgUnit) throws NotSuccessException {
        JobRequest jr = wc.newRepository(repo, orgUnit);
        checkJobStatus(jr.getJobId());
        return jr;
    }

    @Override
    public Collection<RepositoryResponse> getRepositories() throws NotSuccessException {
        return wc.getRepositories();
    }

    private void sleep(int seconds) {
        try {
            Thread.sleep(1000 * seconds);
        } catch (InterruptedException iex) {
            LOG.error("Thread sleep interrupted.", iex);
        }
    }
}
